use crate::model::Modelable;

/// A database entity, aka a struct representing a row in a table
pub trait Entity: 'static + for<'de> serde::Deserialize<'de> + serde::Serialize {
    type ID: EntityID;
    type IDColumn: EntityColumn<Entity = Self>;
    fn table_name() -> &'static str;
    fn column_count() -> usize
    where
        Self: Sized;
    fn columns() -> &'static [&'static dyn EntityColumn<Entity = Self>];
    fn id_column() -> Self::IDColumn;

    fn visit_values<E, F: FnMut(&dyn Modelable) -> Result<(), E>>(
        &self,
        visit: &mut F,
    ) -> Result<(), E>;

    fn build_from(stmt: &sqlite::Statement) -> sqlite::Result<Self>
    where
        Self: Sized;
}

/// Trait representing the columns of a database entity
pub trait EntityColumn: 'static + Send + Sync {
    type Entity: Entity;

    fn table_name(&self) -> &'static str {
        <Self::Entity as Entity>::table_name()
    }

    fn column_typeid(&self) -> std::any::TypeId {
        std::any::TypeId::of::<Self>()
    }

    fn sql_type(&self) -> &'static str;
    fn index(&self) -> usize;
    fn name(&self) -> &'static str;

    fn fk_table_name(&self) -> Option<&'static str>;
    fn fk_column_name(&self) -> Option<&'static str>;
}

/// Trait for entity IDs in the database
pub trait EntityID: 'static + std::fmt::Debug + Copy + Modelable {
    type Entity: Entity;
    fn from_raw_id(raw: i64) -> Self;
    fn raw_id(&self) -> i64;
}

/// Trait for an index over a column
pub trait Index {
    fn index_name() -> &'static str
    where
        Self: Sized;
    fn table_name() -> &'static str
    where
        Self: Sized;
    fn column_names() -> &'static [&'static str]
    where
        Self: Sized;
    fn unique() -> bool
    where
        Self: Sized;
}
