use crate::Error;
use std::hash::{Hash, Hasher};

use crate::entity::{Entity, EntityID};

pub mod build;
pub mod delete;
pub mod filter;
pub mod resolve;
pub mod select;
pub mod update;

pub use filter::Filterable;
pub use resolve::Resolvable;
pub use update::Settable;

/// Wraps an entity with its ID, for example as a query result.
///
/// The wrapped value is accessible via `Deref`, so this should be mostly
/// transparent.
#[derive(Debug)]
pub struct WithID<T: Entity> {
    wrap: T,
    id: <T as Entity>::ID,
}

impl<T: Entity> WithID<T> {
    pub fn new(wrap: T, id: <T as Entity>::ID) -> Self {
        Self { wrap, id }
    }

    fn wrap(what: T, raw_id: i64) -> Self {
        Self {
            wrap: what,
            id: <T as Entity>::ID::from_raw_id(raw_id),
        }
    }

    pub fn wrapped(self) -> T {
        self.wrap
    }
}

impl<T: Entity> WithID<T> {
    pub fn id(&self) -> <T as Entity>::ID {
        self.id
    }
}

impl<T: Entity> AsRef<T> for WithID<T> {
    fn as_ref(&self) -> &T {
        &self.wrap
    }
}

impl<T: Entity> std::ops::Deref for WithID<T> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        &self.wrap
    }
}

impl<T: Entity> std::ops::DerefMut for WithID<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.wrap
    }
}

type CacheIndex = u64;

/// The query interface for a database.
///
/// As the query interface provides some level of caching, try to strive for as much sharing as
/// possible. Passing around `QueryInterface` references instead of `DB` references is a good way
/// to achieve this. However, `QueryInterface` is explicitly `!Send`, so you may need to use
/// something like a [`DBPool`](../struct.DBPool.html) to acquire instances locally if you're using
/// threads or asynchronous programming.
pub struct QueryInterface<'l> {
    db: &'l crate::DB,

    cache: std::cell::RefCell<std::collections::HashMap<CacheIndex, sqlite::Statement<'l>>>,

    // use a phantom non-Send-able type to implement !Send for QueryInterface
    prevent_send: std::marker::PhantomData<*mut ()>,
}

impl<'l> QueryInterface<'l> {
    /// Creates a QueryInterface. Generally you probably want to be using
    /// `DBPool::query_interface()`.
    pub fn new(db: &'l crate::DB) -> Self {
        Self {
            db,
            cache: std::cell::RefCell::new(std::collections::HashMap::new()),
            prevent_send: std::marker::PhantomData,
        }
    }

    /// Helper function to process an expected single result, discarding the rest
    fn expect_one_result<T>(
        &self,
        stmt: &mut sqlite::Statement,
        with_result: &mut dyn FnMut(&mut sqlite::Statement) -> Result<T, crate::Error>,
    ) -> Result<Option<T>, crate::Error> {
        let state = stmt.next()?;
        if state != sqlite::State::Row {
            return Ok(None);
        }

        Ok(Some(with_result(stmt)?))
    }

    /// Helper function to process an expected zero results
    /// Note that this errors out if there is any result
    fn expect_no_result(&self, stmt: &mut sqlite::Statement) -> Result<(), crate::Error> {
        let state = stmt.next()?;
        if state != sqlite::State::Done {
            return Err(crate::Error::ExecFailure);
        }

        Ok(())
    }
}

impl<'l> QueryInterface<'l> {
    fn with_cache<
        Return,
        Create: Fn() -> sqlite::Statement<'l>,
        With: FnMut(&mut sqlite::Statement<'l>) -> Return,
    >(
        &self,
        hash: u64,
        create: Create,
        mut with: With,
    ) -> Return
where {
        let mut cache = self.cache.borrow_mut();
        let query = cache.entry(hash).or_insert_with(create);
        let res = with(query);
        query.reset().expect("Couldn't reset query");
        res
    }
}

impl<'l> QueryInterface<'l> {
    /// Add an entity to its table in the database. Alias for `add`.
    pub fn insert<T: Entity + serde::Serialize>(
        &self,
        m: &T,
    ) -> Result<<T as Entity>::ID, crate::Error> {
        self.add(m)
    }

    /// Add an entity to its table in the database.
    pub fn add<T: Entity + serde::Serialize>(
        &self,
        m: &T,
    ) -> Result<<T as Entity>::ID, crate::Error> {
        let mut hasher = std::collections::hash_map::DefaultHasher::new();
        "add".hash(&mut hasher);
        std::any::TypeId::of::<T>().hash(&mut hasher);
        self.with_cache(
            hasher.finish(),
            &|| {
                let placeholders = (0..(<T as Entity>::column_count() - 1))
                    .map(|_| "?".to_string())
                    .collect::<Vec<_>>()
                    .join(",");

                self.db
                    .conn
                    .prepare(&format!(
                        "INSERT INTO \"{}\" VALUES (NULL, {}) RETURNING \"id\"",
                        <T as Entity>::table_name(),
                        placeholders
                    ))
                    .expect("")
            },
            &mut |stmt: &mut sqlite::Statement<'_>| {
                crate::model::store::serialize_into(stmt, m)?;

                let rowid = self.expect_one_result(stmt, &mut |stmt| {
                    stmt.read::<i64>(0).map_err(crate::Error::DBError)
                })?;

                Ok(<T as Entity>::ID::from_raw_id(rowid.unwrap()))
            },
        )
    }
}

// General query interface
impl<'l> QueryInterface<'l> {
    /// Get an entity from its table, aka perform a `SELECT` query.
    pub fn get<'a, 'b, T: Entity>(&'a self) -> select::Select<'b, 'l, T>
    where
        'a: 'b,
    {
        select::Select::new(self)
    }

    /// Update an entity in its table, aka perform an `UPDATE` query, by selecting columns.
    pub fn update<'a, 'b, T: Entity>(&'a self) -> update::Update<'b, 'l, T>
    where
        'a: 'b,
    {
        update::Update::new(self)
    }

    /// Delete an entity from its table, aka perform a DELETE query.
    pub fn delete<'a, 'b, T: Entity>(&'a self) -> delete::Delete<'b, 'l, T>
    where
        'a: 'b,
    {
        delete::Delete::new(self)
    }
}

impl<'l> QueryInterface<'l> {
    /// Update an entity in its table, aka perform an `UPDATE` query, for all columns.
    pub fn put<T: Entity>(&self, what: &WithID<T>) -> Result<(), Error> {
        self.update().to(what.as_ref()).by(T::id_column(), &what.id()).exec()
    }
}

#[cfg(test)]
mod test_build {
    use microrm_macros::Entity;
    use serde::{Deserialize, Serialize};

    #[derive(Entity, Serialize, Deserialize)]
    #[microrm_internal]
    pub struct KVStore {
        key: String,
        value: String,
    }

    #[test]
    fn simple_get() {
        use super::*;
        let db = crate::DB::new_in_memory(crate::Schema::new().entity::<KVStore>()).unwrap();
        let qi = db.query_interface();

        assert!(qi.get().by(KVStore::Key, "abc").result().is_ok());
    }
}
