//! Static in-place query construction.
//!
//! Interface:
//! - `qi.get().by(KVStore::Key, &key).result()`

use crate::Error;
use std::{
    collections::HashMap,
    hash::{Hash, Hasher},
};

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum QueryPart {
    Root,
    Set,
    Where,
}

#[derive(Debug)]
pub struct DerivedQuery(HashMap<QueryPart, Vec<String>>);

impl DerivedQuery {
    pub(crate) fn new() -> Self {
        Self(HashMap::new())
    }

    pub(crate) fn add(mut self, to: QueryPart, what: String) -> Self {
        self.0.entry(to).or_insert_with(Vec::new).push(what);
        self
    }

    pub(crate) fn assemble(mut self) -> String {
        let root = self.0.remove(&QueryPart::Root).unwrap().remove(0);

        let set_ = match self.0.remove(&QueryPart::Set) {
            None => String::new(),
            Some(v) => {
                format!(
                    "SET {}",
                    v.into_iter()
                        .reduce(|a, b| format!("{}, {}", a, b))
                        .unwrap()
                )
            }
        };

        let where_ = match self.0.remove(&QueryPart::Where) {
            None => String::new(),
            Some(v) => {
                format!(
                    "WHERE {}",
                    v.into_iter()
                        .reduce(|a, b| format!("{} AND {}", a, b))
                        .unwrap()
                )
            }
        };

        log::trace!("built SQL query: {} {} {}", root, set_, where_);

        format!("{} {} {}", root, set_, where_)
    }
}

pub trait StaticVersion {
    type Is: 'static;

    fn type_id() -> std::any::TypeId {
        std::any::TypeId::of::<Self::Is>()
    }
}

/// Any query component
pub trait QueryComponent: StaticVersion {
    fn derive(&self) -> DerivedQuery {
        DerivedQuery::new()
    }

    fn contribute<H: Hasher>(&self, hasher: &mut H);

    /// returns the next index to use for binding
    fn bind(&self, stmt: &mut sqlite::Statement<'_>) -> Result<usize, Error>;
}

#[derive(Debug)]
pub enum CompareOp {
    LessThan,
    AtMost,
    Equals,
    AtLeast,
    MoreThan,
}

impl CompareOp {
    pub(crate) fn ch(&self) -> &'static str {
        match self {
            Self::LessThan => "<",
            Self::AtMost => "<=",
            Self::Equals => "=",
            Self::AtLeast => ">=",
            Self::MoreThan => ">",
        }
    }
}
